module Mucks
  extend self

  module SystemIO
    extend self

    def load_yaml(*args)
      YAML.load_file(*args)
    end
  end

  class BaseConfig
    attr_reader :config
    def initialize(path)
      @config = SystemIO.load_yaml(File.expand_path(path))
      show_self if Config.verbose
    end

    def show_self
      Output.wrap_with(self.class) do
        s = ''
        @config.each do |key, value|
          s += "#{key} => "
          if value.is_a?(Array) then
            s += stringify_elements_for(value)
          else
            s += "\n"
            s += "  #{value}}\n"
          end
        end
        puts s
      end
    end

    def stringify_elements_for(array)
      s = "[\n"
      array.each do |element|
        s += "  #{element.inspect},\n"
      end
      s += "]\n"
    end
  end

  class LayoutConfig < BaseConfig
    def initialize(path)
      super
      config = {}
      @config.each do |k, v|
        config[k] = get_commands_for(v).flatten
      end
      @config = config
    end

    def layout_keys
      @config.keys
    end

    def [](key)
      @config[key]
    end

    private

    def get_commands_for(list)
      list.map do |item|
        candidate = @config[item]
        if candidate
          get_commands_for candidate
        else
          item
        end
      end
    end
  end

  class SessionConfig < BaseConfig
    attr_reader :sessions, :session_names

    def initialize(path)
      super

      @sessions = unique_sessions.map do |session_name|
        SessionElement.new session_name, @config[session_name]
      end

      @session_names = Set.new
      @config.each do |k, v|
        @session_names << k
      end
    end

    def layouts
      @sessions.map do |session|
        session.layout
      end.uniq
    end

    def remove_by_layout(layout_name)
      @sessions.reject! do |session|
        session.layout == layout_name
      end
    end

    private

    def unique_sessions
      Config.sessions.map do |session_name|
        get_sessions_for(session_name)
      end.flatten.uniq.compact
    end

    def get_sessions_for(session_name)
      session_config = @config[session_name]
      case session_config
      when Array
        session_config.map do |sess_name|
          get_sessions_for sess_name
        end
      when Hash
        session_name
      else
        puts "Invalid session selected: #{session_name}"
        nil
      end
    end
  end

  class SessionElement
    attr_reader :name, :path, :layout

    def initialize(name, session)
      @name   = name
      @path   = File.expand_path(session['path'])
      @layout = session['layout']
    end
  end

  module CLI
    extend self

    def run(args)
      Config.configure(args)
      Main.run
    end
  end

  module Main
    extend self

    def run
      set_configs
      validate_configs

      if Config.list
        show_all_configured_sessions
        return
      end

      return if @session_config.sessions.empty?

      show_valid_sessions
      invoke_sessions

      attach @session_config.sessions.last if Config.attach
    end

    private

    def show_all_configured_sessions
      @session_config.show_self
    end

    def invoke_sessions
      @session_config.sessions.each do |session|
        invoke session
      end
    end

    def invoke(session)
      Output.wrap_with("#{session.name} Session", Config.dry_run) do
        if start(session) == 0 || Config.dry_run
          configure session
        end
      end
    end

    def set_configs
      @session_config = SessionConfig.new('~/.mucks/sessions.yaml')
      @layout_config  = LayoutConfig.new('~/.mucks/layouts.yaml')
    end

    def show_valid_sessions
      Output.wrap_with('Valid Sessions', Config.dry_run) do
        if Config.verbose || Config.dry_run
          puts @session_config.sessions.map(&:name)
        end
      end
    end

    def attach(session)
      Executor.run "tmux attach -t #{session.name}"
    end

    def start(session)
      Dir.chdir session.path
      Executor.run "tmux new -s #{session.name} -d"
    end

    def configure(session)
      layout = layout_for(session)
      layout.each do |command|
        Executor.run command, session
      end
    end

    def layout_for(session)
      layout = []
      @layout_config[session.layout].each do |command|
        if set_path?(command)
          layout << parse_path(session, command)
        else
          layout << command
        end
      end
      layout
    end

    def set_path?(command)
      command.downcase == 'set_path' || command.downcase =~ /^set_path /
    end

    def parse_path(session, command)
      path = command.gsub(/set_path */i, '')

      if path.empty?
        path = session.path
      else
        path = File.expand_path(path)
      end

      tmux_chdir path
    end

    def tmux_chdir(path)
      "tmux send-keys 'cd \"#{path}\"' C-m"
    end

    def validate_configs
      invalid = []
      @session_config.layouts.each do |layout|
        unless @layout_config.layout_keys.include?(layout)
          puts "Invalid layout selected: #{layout}"
          invalid << layout
        end
      end

      invalid.each do |layout|
        @session_config.remove_by_layout layout
      end
    end
  end

  module Config
    extend self

    attr_reader :attach, :verbose, :dry_run, :list, :sessions

    def configure(args)
      @params, @sessions = args.partition { |a| a.match(/^-/) }
      @params = @params.reduce('') do |memo, param|
        param = param.gsub(/-/, '')
        memo + param
      end

      set_basic_attributes

      show if self.verbose
    end

    private

    def set_basic_attributes
      @attach = !check_for('d')
      @verbose = check_for('v')
      @dry_run = check_for('n')
      @list    = check_for('ls')
    end

    def show
      Output.wrap_with('Config') do
        [:attach, :verbose, :dry_run, :list, :sessions].each do |attr|
          puts "#{attr} = #{self.send attr}"
        end
      end
    end

    def check_for(flag)
      return unless @params
      @params.include?(flag)
    end
  end

  module Executor
    extend self
    def run(command, session = nil)
      puts command if Config.verbose || Config.dry_run
      `#{command}` unless Config.dry_run
      $?
    end
  end

  module Output
    extend self
    def wrap_with(text, show = true, &block)
      if block_given?
        puts "*** #{text} Start ".ljust(80, '*') if Config.verbose || show
        yield
        puts "--- #{text} End ".ljust(80, '-') if Config.verbose || show
      else
        puts "--- #{text} ".ljust(80, '-') if Config.verbose || show
      end
    end
  end

  def muckity_muck(args)
    CLI.run(args)
  end
end
